"""List files office plugin"""

from collections import OrderedDict
from collections.abc import Sequence
from typing import Any

from cmem_plugin_base.dataintegration.context import ExecutionContext, ExecutionReport
from cmem_plugin_base.dataintegration.description import Icon, Plugin, PluginAction, PluginParameter
from cmem_plugin_base.dataintegration.entity import Entities, Entity, EntityPath, EntitySchema
from cmem_plugin_base.dataintegration.parameter.choice import ChoiceParameterType
from cmem_plugin_base.dataintegration.parameter.password import Password, PasswordParameterType
from cmem_plugin_base.dataintegration.plugins import WorkflowPlugin
from cmem_plugin_base.dataintegration.ports import FixedNumberOfInputs, FixedSchemaPort
from office365.graph_client import GraphClient
from office365.runtime.client_object import ClientObject

from cmem_plugin_office365.autocompletion import (
    DriveDirectoryParameterType,
    FolderDirectoryParameterType,
    ResourceDirectoryParameterType,
)
from cmem_plugin_office365.retrieval import OfficeRetrieval

RESOURCE_CHOICE = OrderedDict({"personal": "Personal", "site": "Site"})
MAX_WORKERS = 32


def generate_schema() -> EntitySchema:
    """Generate the schema for entities"""
    return EntitySchema(
        type_uri="",
        paths=[
            EntityPath(path="type"),
            EntityPath(path="name"),
            EntityPath(path="web_url"),
            EntityPath(path="parent_reference"),
            EntityPath(path="microsoft_graph_download_url"),
            EntityPath(path="created_by"),
            EntityPath(path="created_date_time"),
            EntityPath(path="eTag"),
            EntityPath(path="id"),
            EntityPath(path="last_modified_by"),
            EntityPath(path="last_modified_date_time"),
            EntityPath(path="cTag"),
            EntityPath(path="size"),
        ],
    )


def setup_max_workers(max_workers: int) -> int:
    """Return the correct amount of workers"""
    if 0 < max_workers <= MAX_WORKERS:
        return max_workers
    raise ValueError("Range of max_workers exceeded")


@Plugin(
    label="List Office 365 Files",
    plugin_id="cmem_plugin_office365-List",
    description="List files from OneDrive or Sites",
    documentation="""
This workflow task creates a structured output from a specified Office 365 instance.
For this to work a registered app in Microsoft's Entra ID space is necessary.
Further information can be found [here](https://learn.microsoft.com/en-us/entra/identity-platform/quickstart-register-app).

After registering an application, it needs to be granted application wide API permissions:
- Files.Read.All
- Sites.Read.All

Admin consent is required to activate these permissions.
With this setup, anyone with the secret can access all users' OneDrives and all Sharepoint/Team
sites.

#### Important
Make sure only trusted admins can create or manage secrets!
Whoever holds the secrets has all the access to granted resources so best not to distribute
recklessly.
    """,
    icon=Icon(package=__package__, file_name="o365-list.svg"),
    actions=[
        PluginAction(
            name="preview_results", label="Preview results", description="Preview the results."
        )
    ],
    parameters=[
        PluginParameter(
            name="tenant_id",
            label="Tenant ID",
            description="ID of your tenant. Can be seen within your registered application",
        ),
        PluginParameter(
            name="client_id",
            label="Client ID",
            description="Client ID of your registered application.",
        ),
        PluginParameter(
            param_type=PasswordParameterType(),
            name="client_secret",
            label="Client secret",
            description="Client secret created withing your registered application.",
        ),
        PluginParameter(
            param_type=ChoiceParameterType(OrderedDict({"personal": "Personal", "site": "Site"})),
            name="type_resource",
            label="Type resource",
            description="The type of resource you want the data to be extracted from. "
            "This can either be a site or a users share",
        ),
        PluginParameter(
            name="target_resource",
            label="Target resource",
            description="Target resource which files will be listed from. This can either be a "
            "specific users share address or a microsoft site URL.",
            param_type=ResourceDirectoryParameterType("resource", "Resource"),
        ),
        PluginParameter(
            name="drives",
            label="Drives",
            description="A list of drives from the selected target resource.",
            param_type=DriveDirectoryParameterType("drives", "Drives"),
        ),
        PluginParameter(
            name="path",
            label="Directory path",
            description="The path of a directory that needs to be transformed. Includes all "
            "subdirectories by default",
            param_type=FolderDirectoryParameterType("path", "Path"),
        ),
        PluginParameter(
            name="regex",
            label="Regular expression",
            description="A regular expression performed on all the files within the selected path",
            default_value="^.*$",
        ),
        PluginParameter(
            name="no_subfolder",
            label="Exclude files in subfolders",
            description="A flag indicating if files should only be listed from subfolders or not.",
        ),
        PluginParameter(
            name="max_workers",
            label="Maximum amount of workers",
            description="Specifies the maximum number of threads used for parallel execution of "
            "the workflow. "
            "The default is 32, and the valid range is 1 to 32. "
            "Note: Due to known throttling limits imposed by Microsoft, running with high "
            "parallelism may cause errors. "
            "If you encounter issues, try reducing the number of threads to 1.",
            advanced=True,
            default_value=32,
        ),
    ],
)
class ListPlugin(WorkflowPlugin):
    """List Plugin Office 365"""

    def __init__(  # noqa: PLR0913
        self,
        tenant_id: str,
        client_id: str,
        client_secret: Password | str,
        type_resource: str,
        target_resource: str,
        drives: str,
        max_workers: int = 32,
        path: str = "",
        regex: str = "",
        no_subfolder: bool = False,
    ) -> None:
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = (
            client_secret if isinstance(client_secret, str) else client_secret.decrypt()
        )
        self.path = path
        self.type_resource = RESOURCE_CHOICE[type_resource]
        self.target_resource = target_resource
        self.drives = drives
        self.regex = rf"{regex}"
        self.no_subfolder = no_subfolder
        self.max_workers = setup_max_workers(max_workers)
        self.client = GraphClient(tenant=tenant_id).with_client_secret(
            self.client_id, self.client_secret
        )
        self.drive = self.create_drive()
        self.input_ports = FixedNumberOfInputs([])
        self.output_port = FixedSchemaPort(schema=generate_schema())

    def preview_results(self) -> str:
        """Preview the result of the execution"""
        files: list[Any] = []
        result = []
        retrieval = OfficeRetrieval(
            drive=self.drive, regex=self.regex, no_subfolder=self.no_subfolder
        )
        files = retrieval.list_files_parallel(
            path=self.path, files=files, context=None, no_of_max_hits=10, workers=self.max_workers
        )
        if files:
            result.append("\nExample files: (max 10)\n")
            for file in files:
                props = file.properties
                result.append("\n" + "- " + props["name"])
            return "".join(result)
        result.append("\nNo files were found with the given instructions\n")
        return "".join(result)

    def create_drive(self) -> ClientObject:
        """Create drive"""
        s = self.client.sites.get_by_url(self.target_resource)
        return s.drive

    def execute(self, inputs: Sequence[Entities], context: ExecutionContext) -> Entities:
        """Execute workflow task"""
        _ = inputs

        context.report.update(
            ExecutionReport(entity_count=0, operation="wait", operation_desc="files listed.")
        )

        files: list[Any] = []
        entities = []
        retrieval = OfficeRetrieval(
            drive=self.drive, regex=self.regex, no_subfolder=self.no_subfolder
        )
        all_files = retrieval.list_files_parallel(
            path=self.path, files=files, context=context, workers=self.max_workers
        )
        for file in all_files:
            try:
                if context.workflow.status() == "Canceling":
                    break
            except AttributeError:
                pass
            props = file.properties
            entity_uri = props["webUrl"]
            values = [
                ["file"],
                [props["name"]],
                [props["webUrl"]],
                [props["parentReference"].name],
                [props["@microsoft.graph.downloadUrl"]],
                [str(props["createdBy"])],
                [props["createdDateTime"].strftime("%Y-%m-%d %H:%M:%S")],
                [props["eTag"]],
                [props["id"]],
                [str(props["lastModifiedBy"])],
                [props["lastModifiedDateTime"].strftime("%Y-%m-%d %H:%M:%S")],
                [props["cTag"]],
                [str(props["size"])],
            ]
            entities.append(Entity(uri=entity_uri, values=values))
            context.report.update(
                ExecutionReport(
                    entity_count=len(entities),
                    operation="write",
                    operation_desc="entities generated",
                )
            )

        context.report.update(
            ExecutionReport(
                entity_count=len(entities),
                operation="done",
                operation_desc="entities generated",
                sample_entities=Entities(entities=iter(entities[:10]), schema=generate_schema()),
            )
        )
        return Entities(entities=iter(entities), schema=generate_schema())



"""Autocompletion"""

from typing import Any, ClassVar

from cmem_plugin_base.dataintegration.context import PluginContext
from cmem_plugin_base.dataintegration.types import Autocompletion, StringParameterType
from office365.graph_client import GraphClient
from office365.runtime.client_object import ClientObject


def list_all_sites(depend_on_parameter_values: list[Any]) -> ClientObject:
    """List all available sites"""
    client = get_office_client(depend_on_parameter_values)
    return client.sites.get().execute_query()


def get_office_client(depend_on_parameter_values: list[Any]) -> GraphClient:
    """Initialize the office client"""
    return GraphClient(tenant=depend_on_parameter_values[0]).with_client_secret(
        client_id=depend_on_parameter_values[1],
        client_secret=depend_on_parameter_values[2]
        if isinstance(depend_on_parameter_values[2], str)
        else depend_on_parameter_values[2].decrypt(),
    )


def sort_suggestions(suggestions: list[Autocompletion], query_terms: list[str]) -> None:
    """Sort autocompleted suggestions"""
    suggestions.sort(
        key=lambda x: (
            not all(term.lower() in x.label.lower() for term in query_terms),
            x.label.lower(),
        )
    )


class ResourceDirectoryParameterType(StringParameterType):
    """Resource autocompletion parameter"""

    def __init__(
        self,
        url_expand: str,
        display_name: str,
    ) -> None:
        self.url_expand = url_expand
        self.display_name = display_name
        self.suggestions: list[Autocompletion] = []

    allow_only_autocompleted_values = True

    autocompletion_depends_on_parameters: ClassVar[list[str]] = [
        "tenant_id",
        "client_id",
        "client_secret",
        "type_resource",
    ]

    def autocomplete(
        self,
        query_terms: list[str],
        depend_on_parameter_values: list[Any],
        context: PluginContext,
    ) -> list[Autocompletion]:
        """Autocomplete resources"""
        _ = context
        entered_term = "".join(query_terms).lower()
        target_type = depend_on_parameter_values[3]
        all_sites = list_all_sites(depend_on_parameter_values)
        autocompletion = []

        for site in all_sites:
            props = site.properties
            if (
                not props["isPersonalSite"]
                and target_type == "site"
                and entered_term in props["webUrl"].lower()
            ):
                autocompletion.append(Autocompletion(value=props["webUrl"], label=props["webUrl"]))
            elif (
                props["isPersonalSite"]
                and target_type == "personal"
                and entered_term in props["webUrl"].lower()
            ):
                url = props["webUrl"]
                label = props["webUrl"]
                autocompletion.append(Autocompletion(value=url, label=label))
        sort_suggestions(autocompletion, query_terms)
        self.suggestions = autocompletion
        return self.suggestions


def add_parent_folder(result: list[Autocompletion], selected_path: str) -> None:
    """Add parent folder of given path to autocompletion list"""
    parent_folder = "/".join(selected_path.rstrip("/").split("/")[:-1])
    if parent_folder:
        parent_folder = f"{parent_folder}/"
        result.append(Autocompletion(value=parent_folder, label=parent_folder))


class FolderDirectoryParameterType(StringParameterType):
    """Folder autocompletion parameter"""

    def __init__(
        self,
        url_expand: str,
        display_name: str,
    ) -> None:
        self.url_expand = url_expand
        self.display_name = display_name
        self.suggestions: list[Autocompletion] = []

    autocompletion_depends_on_parameters: ClassVar[list[str]] = [
        "tenant_id",
        "client_id",
        "client_secret",
        "target_resource",
        "path",
        "drives",
    ]

    def autocomplete(
        self,
        query_terms: list[str],
        depend_on_parameter_values: list[Any],
        context: PluginContext,
    ) -> list[Autocompletion]:
        """Autocompletion for folder"""
        _ = context

        entered_directory = "".join(query_terms)
        selected_path = depend_on_parameter_values[4]
        result: list[Autocompletion] = []

        if selected_path == "":
            client = get_office_client(depend_on_parameter_values)
            drive = client.drives[depend_on_parameter_values[5]].get().execute_query()
            drive_items = drive.root.children.get().execute_query()
            for item in drive_items:
                props = item.properties
                if "folder" in props:
                    result.append(
                        Autocompletion(value=props["name"] + "/", label=props["name"] + "/")
                    )
            self.suggestions = result
            return self.suggestions

        if selected_path != "" and entered_directory == "":
            return self.suggestions

        if entered_directory[-1] == "/" and selected_path != "":
            client = get_office_client(depend_on_parameter_values)
            drive = client.drives[depend_on_parameter_values[5]].get().execute_query()
            drive_items = drive.root.get_by_path(selected_path[:-1]).children.get().execute_query()
            for item in drive_items:
                props = item.properties
                if "folder" in props:
                    result.append(
                        Autocompletion(
                            value=selected_path + props["name"] + "/",
                            label=selected_path + props["name"] + "/",
                        )
                    )

            add_parent_folder(result, selected_path)

            self.suggestions = result

            for suggestion in self.suggestions:
                if suggestion.label == "/":
                    return self.suggestions

            self.suggestions.append(Autocompletion(value="", label="/"))

            return self.suggestions

        return self.suggestions


class DriveDirectoryParameterType(StringParameterType):
    """Drive autocompletion parameter"""

    def __init__(
        self,
        url_expand: str,
        display_name: str,
    ) -> None:
        self.url_expand = url_expand
        self.display_name = display_name
        self.suggestions: list[Autocompletion] = []

    autocompletion_depends_on_parameters: ClassVar[list[str]] = [
        "tenant_id",
        "client_id",
        "client_secret",
        "target_resource",
        "path",
    ]

    def autocomplete(
        self,
        query_terms: list[str],
        depend_on_parameter_values: list[Any],
        context: PluginContext,
    ) -> list[Autocompletion]:
        """Autocomplete the drives"""
        _ = query_terms
        _ = context
        client = get_office_client(depend_on_parameter_values)
        target = depend_on_parameter_values[3]
        drives = client.sites.get_by_url(target).drives.get().execute_query()
        return [Autocompletion(value=drive.id, label=drive.name) for drive in drives]



"""Parallel directory retrieval"""

import re
import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any

from cmem_plugin_base.dataintegration.context import ExecutionContext, ExecutionReport
from office365.runtime.client_object import ClientObject


def context_report(context: ExecutionContext, files: list[Any]) -> None:
    """Report for user context"""
    if context is not None:
        context.report.update(
            ExecutionReport(
                entity_count=len(files), operation="wait", operation_desc="files listed"
            )
        )


def _strip_path(path: str) -> str:
    """Strip path of trailing '/'"""
    if path[-1] == "/":
        path = path[:-1]
    return path


class OfficeRetrieval:
    """Retrieval class for Office 365 folders and files"""

    def __init__(self, drive: ClientObject, no_subfolder: bool, regex: str) -> None:
        self.drive = drive
        self.no_subfolder = no_subfolder
        self.regex = regex
        self.lock = threading.Lock()
        self.stop_event = threading.Event()

    def list_files_parallel(  # noqa: PLR0913
        self,
        path: str,
        files: list[Any],
        context: ExecutionContext | None,
        depth: int = -1,
        curr_depth: int = 0,
        no_of_max_hits: int = -1,
        workers: int = 32,
    ) -> list[Any]:
        """List files"""
        if curr_depth == 0:
            self.stop_event.clear()
        self.cancel_listdir(context)
        if self.stop_event.is_set() or (depth != -1 and curr_depth >= depth):
            return files

        subdirectories: list[str] = []
        items = self._get_folder_items(path)

        for item in items:
            self.cancel_listdir(context)
            if self.stop_event.is_set():
                return files

            added = self.add_node(files, item, no_of_max_hits)
            context_report(context, files)

            if added and self.check_stop(files, no_of_max_hits):
                return files

            if item.is_folder and not self.no_subfolder:
                subdirectories.append(path + "/" + item.name)

        if not self.stop_event.is_set() and subdirectories:
            with ThreadPoolExecutor(max_workers=workers) as executor:
                futures = [
                    executor.submit(
                        self.list_files_parallel,
                        sd,
                        files,
                        None,
                        depth,
                        curr_depth + 1,
                        no_of_max_hits,
                        workers,
                    )
                    for sd in subdirectories
                ]
                for fut in as_completed(futures):
                    self.cancel_listdir(context)
                    if self.stop_event.is_set():
                        break
                    fut.result()
                    context_report(context, files)

        context_report(context, files)
        return files

    def cancel_listdir(self, context: ExecutionContext) -> None:
        """Cancel listdir if workflow is cancelled"""
        try:
            if context.workflow.status() == "Canceling":
                self.stop_event.set()
        except AttributeError:
            pass

    def add_node(self, files: list[Any], item: Any, no_of_max_hits: int) -> bool:  # noqa: ANN401
        """Add file or folder node to result"""
        with self.lock:
            if no_of_max_hits != -1 and len(files) >= no_of_max_hits:
                self.stop_event.set()
                return False

            props = item.properties
            if (re.fullmatch(self.regex, props["name"])) and not item.is_folder:
                files.append(item)
                if no_of_max_hits != -1 and len(files) >= no_of_max_hits:
                    self.stop_event.set()
                return True

        return False

    def check_stop(self, files: list[Any], max_results: int) -> bool:
        """Check whether max_results reached and stop if so"""
        with self.lock:
            if max_results != -1 and len(files) >= max_results:
                self.stop_event.set()
                return True
        return False

    def _get_folder_items(self, path: str) -> Any:  # noqa: ANN401
        """Get the items within a folder"""
        if path == "":
            return self.drive.root.children.get().execute_query()
        path = _strip_path(path)
        return self.drive.root.get_by_path(path).children.get().execute_query()



"""Upload workflow task for Office 365"""

import mimetypes
import tempfile
from collections import OrderedDict
from collections.abc import Sequence
from pathlib import Path

from cmem.cmempy.workspace.projects.resources.resource import get_resource_response
from cmem_plugin_base.dataintegration.context import ExecutionContext, ExecutionReport
from cmem_plugin_base.dataintegration.description import Icon, Plugin, PluginParameter
from cmem_plugin_base.dataintegration.entity import Entities
from cmem_plugin_base.dataintegration.parameter.choice import ChoiceParameterType
from cmem_plugin_base.dataintegration.parameter.password import Password, PasswordParameterType
from cmem_plugin_base.dataintegration.parameter.resource import ResourceParameterType
from cmem_plugin_base.dataintegration.plugins import WorkflowPlugin
from cmem_plugin_base.dataintegration.ports import FixedNumberOfInputs, FixedSchemaPort, InputPorts
from cmem_plugin_base.dataintegration.typed_entities.file import (
    File,
    FileEntitySchema,
    LocalFile,
    ProjectFile,
)
from cmem_plugin_base.dataintegration.utils import setup_cmempy_user_access
from office365.graph_client import GraphClient
from office365.runtime.client_object import ClientObject

from cmem_plugin_office365.autocompletion import (
    DriveDirectoryParameterType,
    FolderDirectoryParameterType,
    ResourceDirectoryParameterType,
)

RESOURCE_CHOICE = OrderedDict({"personal": "Personal", "site": "Site"})
MAX_WORKERS = 32


@Plugin(
    label="Office 365 Upload Files",
    description="Upload files to OneDrive or a site Sharepoint",
    documentation="""
This workflow task upload files to specified Office 365 instance.
For this to work a registered app in Microsoft's Entra ID space is necessary.
Further information can be found [here](https://learn.microsoft.com/en-us/entra/identity-platform/quickstart-register-app).

After registering an application, it needs to be granted application wide API permissions:
- Files.Read.All, Files.Write.All
- Sites.Read.All, Sites.Write.All

Admin consent is required to activate these permissions.
With this setup, anyone with the secret can access all users' OneDrives and all Sharepoint/Team
sites.

#### Important
Make sure only trusted admins can create or manage secrets!
Whoever holds the secrets has all the access to granted resources so best not to distribute
recklessly.
    """,
    icon=Icon(package=__package__, file_name="o365-upload.svg"),
    parameters=[
        PluginParameter(
            name="tenant_id",
            label="Tenant ID",
            description="ID of your tenant. Can be seen within your registered application",
        ),
        PluginParameter(
            name="client_id",
            label="Client ID",
            description="Client ID of your registered application.",
        ),
        PluginParameter(
            param_type=PasswordParameterType(),
            name="client_secret",
            label="Client secret",
            description="Client secret created withing your registered application.",
        ),
        PluginParameter(
            param_type=ChoiceParameterType(OrderedDict({"personal": "Personal", "site": "Site"})),
            name="type_resource",
            label="Type resource",
            description="The type of resource you want the data to be extracted from. "
            "This can either be a site or a users share",
        ),
        PluginParameter(
            name="target_resource",
            label="Target resource",
            description="Target resource which files will be listed from. This can either be a "
            "specific users share address or a microsoft site URL.",
            param_type=ResourceDirectoryParameterType("resource", "Resource"),
        ),
        PluginParameter(
            name="drives",
            label="Drives",
            description="A list of drives from the selected target resource.",
            param_type=DriveDirectoryParameterType("drives", "Drives"),
        ),
        PluginParameter(
            name="path",
            label="Directory path",
            description="The path of a directory that needs to be transformed. Includes all "
            "subdirectories by default",
            param_type=FolderDirectoryParameterType("path", "Path"),
        ),
        PluginParameter(
            name="source_file",
            label="File",
            description="The file which you would like to upload. If this is set, no connections "
            "in the workflow are allowed",
            param_type=ResourceParameterType(),
            advanced=True,
            default_value="",
        ),
        PluginParameter(
            name="max_workers",
            label="Maximum amount of workers",
            description="Specifies the maximum number of threads used for parallel execution of "
            "the workflow. "
            "The default is 32, and the valid range is 1 to 32. "
            "Note: Due to known throttling limits imposed by Microsoft, running with high "
            "parallelism may cause errors. "
            "If you encounter issues, try reducing the number of threads to 1.",
            advanced=True,
            default_value=32,
        ),
    ],
)
class UploadPlugin(WorkflowPlugin):
    """Upload plugin Office 365"""

    def __init__(  # noqa: PLR0913
        self,
        tenant_id: str,
        client_id: str,
        client_secret: Password | str,
        type_resource: str,
        target_resource: str,
        drives: str,
        source_file: str,
        max_workers: int = 32,
        path: str = "",
    ) -> None:
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = (
            client_secret if isinstance(client_secret, str) else client_secret.decrypt()
        )
        self.path = path
        self.type_resource = RESOURCE_CHOICE[type_resource]
        self.target_resource = target_resource
        self.drives = drives
        self.source_file = source_file
        self.max_workers = setup_max_workers(max_workers)
        self.client = GraphClient(tenant=tenant_id).with_client_secret(
            self.client_id, self.client_secret
        )
        self.drive = self.create_drive()
        self.input_ports = self.initialise_input_ports()
        self.output_port = None

    def initialise_input_ports(self) -> InputPorts:
        """Initialise the input ports depending on weather a source file is given or not"""
        return (
            FixedNumberOfInputs([FixedSchemaPort(schema=FileEntitySchema())])
            if self.source_file == ""
            else FixedNumberOfInputs([])
        )

    def create_drive(self) -> ClientObject:
        """Create drive"""
        s = self.client.sites.get_by_url(self.target_resource)
        return s.drive

    def execute(self, inputs: Sequence[Entities], context: ExecutionContext) -> Entities:
        """Execute the workflow task"""
        schema = FileEntitySchema()
        files: list[File] = []

        if len(inputs) == 0:
            self.upload_no_input(context, files)
        else:
            self.upload_with_input(context, files, inputs)

        entities = [schema.to_entity(file) for file in files]

        context.report.update(
            ExecutionReport(
                entity_count=len(entities),
                operation="write",
                operation_desc="files uploaded",
                sample_entities=Entities(entities=iter(entities[:10]), schema=schema),
            )
        )

        return Entities(entities=iter(entities), schema=schema)

    def upload_no_input(self, context: ExecutionContext, files: list[File]) -> None:
        """Upload of a source file with no input given"""
        if self.source_file != "":
            setup_cmempy_user_access(context.user)
            file = get_resource_response(context.task.project_id(), self.source_file)
            mime_type, _ = mimetypes.guess_type(self.source_file)
            context.report.update(
                ExecutionReport(
                    entity_count=len(files),
                    operation="upload",
                    operation_desc=f"uploading {self.source_file}",
                )
            )
            with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                tmp_file.write(file.content)
                tmp_file.seek(0)
                with Path(tmp_file.name).open("rb") as f:
                    self.drive.root.upload_file(f).execute_query()

            files.append(LocalFile(path=str(self.source_file), mime=mime_type))

    def upload_with_input(
        self, context: ExecutionContext, files: list[File], inputs: Sequence[Entities]
    ) -> None:
        """Upload of files from a given input"""
        schema = FileEntitySchema()
        for entity in inputs[0].entities:
            try:
                if context.workflow.status() == "Canceling":
                    break
            except AttributeError:
                pass

            entity_file = schema.from_entity(entity)
            file_path = Path(entity_file.path)

            context.report.update(
                ExecutionReport(
                    entity_count=len(files),
                    operation="upload",
                    operation_desc=f"uploading {file_path}",
                )
            )

            if entity_file.file_type != "Local":
                setup_cmempy_user_access(context.user)
                file = get_resource_response(context.task.project_id(), str(file_path))
                with tempfile.TemporaryDirectory() as tmp_dir:
                    tmp_file_path = Path(tmp_dir) / file_path.name
                    tmp_file_path.write_bytes(file.content)
                    with tmp_file_path.open("rb") as f:
                        if self.path == "":
                            self.drive.root.upload_file(f).execute_query()
                        else:
                            self.path = _strip_path(self.path)
                            self.drive.root.get_by_path(self.path).upload_file(f).execute_query()

            elif self.path == "":
                self.drive.root.upload_file(file_path).execute_query()
            else:
                self.path = _strip_path(self.path)
                self.drive.root.get_by_path(self.path).upload_file(file_path).execute_query()

            if entity_file.file_type == "Local":
                files.append(LocalFile(path=str(file_path), mime=entity_file.mime))
            elif entity_file.file_type == "Project":
                files.append(ProjectFile(path=str(file_path), mime=entity_file.mime))


def setup_max_workers(max_workers: int) -> int:
    """Return the correct number of workers"""
    if 0 < max_workers <= MAX_WORKERS:
        return max_workers
    raise ValueError("Range of max_workers exceeded")


def _strip_path(path: str) -> str:
    """Strip path of trailing '/'"""
    if path[-1] == "/":
        path = path[:-1]
    return path
